# %%
import math
import torch

from pymeten import triton_benchmark_sizes, MetricConfig, MetricUnit

from lmetric.llm_quant.gptq_quant_linear import matmul248

import bitsandbytes as bnb
from bitsandbytes.nn import Int8Params
torch.manual_seed(0)


# %%

def get_bnb_weight_state(weight, threshold=6.0, has_fp16_weights=False, memory_efficient_backward=False):
    """
    Constructs a state and a bnb_weight dictionary.

    Args:
        weight (Tensor): The weight tensor.
        threshold (float, optional): Threshold value. Defaults to 6.0.
        has_fp16_weights (bool, optional): Flag indicating if the weights are in fp16 format. Defaults to False.
        memory_efficient_backward (bool, optional): Flag indicating if memory-efficient backward is enabled. Defaults to False.

    Returns:
        dict: A dictionary containing the 'bnb_weight' and 'state'.
    """

    bnb_weight = Int8Params(weight.T, has_fp16_weights=has_fp16_weights, requires_grad=has_fp16_weights).cuda(device='cuda')

    state = bnb.MatmulLtState()
    state.threshold = threshold
    state.has_fp16_weights = has_fp16_weights
    state.memory_efficient_backward = memory_efficient_backward
    state.is_training = has_fp16_weights
    if threshold > 0.0 and not has_fp16_weights:
        state.use_pool = True
    state.CB = bnb_weight.CB
    state.SCB = bnb_weight.SCB

    bnb_weight.CB = None
    bnb_weight.SCB = None

    return {'bnb_weight': bnb_weight, 'state': state}

def lmetric_quant_matmul_Lx3HxH(seqlen=128, hidden_sizes=[2048 * i for i in range(1, 8)], dtype=torch.float16, metric='TFLOPS',
                                kernels=['torch', 'gptq', 'bnb_infer'],
                                bits=4, groupsize=128):
    """
    Performs a quantized matrix multiplication benchmarking with different kernels.

    Args:
        seqlen (int, optional): Sequence length. Defaults to 128.
        hidden_sizes (list, optional): List of hidden sizes. Defaults to [2048 * i for i in range(1, 8)].
        dtype (torch.dtype, optional): Data type. Defaults to torch.float16.
        metric (str, optional): Metric for benchmarking. Defaults to 'TFLOPS'. options: TFLOPS, TBPS, MS
        kernels (list, optional): List of kernels to use. Defaults to ['torch', 'gptq', 'bnb_infer'].
        bits (int, optional): Number of bits for quantization. Defaults to 4.
        groupsize (int, optional): Group size for quantization. Defaults to 128.

    Returns:
        None
    """
    
    maxq = 2 ** bits - 1
    dtype_bytes = torch.finfo(dtype).bits // 8

    if metric == 'TFLOPS':
        metric_config=MetricConfig(
            MetricUnit.TFLOPS,
            calc_measure_fn=lambda size, ms: dtype_bytes*size*size*size / ms)
    elif metric == 'TBPS':
        metric_config=MetricConfig(
            MetricUnit.TBPS,
            calc_measure_fn=lambda size, ms: 3*dtype_bytes*size*size / ms)
    else:
        metric_config=MetricConfig(
            MetricUnit.MS)

    inputs = {
        'a' : lambda size:
            torch.randn((seqlen, 3*size), device='cuda', dtype=dtype),
        'weight': lambda size:
            torch.randn((3*size, size), device='cuda', dtype=dtype),
        'qweight': lambda size:
            torch.zeros((3*size // 32 * bits, size), device='cuda', dtype=torch.int32),
        'qzeros': lambda size:
            torch.zeros((math.ceil(3*size / groupsize), size // 32 * bits),
                        device='cuda', dtype=torch.int32),
        'scales': lambda size:
            torch.zeros((math.ceil(3*size / groupsize), size),
                        device='cuda', dtype=dtype),
        'g_idx': lambda size:
            torch.tensor([i // groupsize for i in range(3*size)],
                        device='cuda', dtype=torch.int32),
        'bnb_dict': lambda size: 
            get_bnb_weight_state(
                torch.randn((3*size, size), device='cuda', dtype=dtype),
            )

    }

    kernelfns = {
        'torch': lambda **kwargs:
            kwargs['a'] @ kwargs['weight'],
        'gptq': lambda **kwargs:
            matmul248(kwargs['a'], kwargs['qweight'], kwargs['scales'], kwargs['qzeros'], kwargs['g_idx'], bits, maxq),
        'bnb_infer': lambda **kwargs:
            bnb.matmul(kwargs['a'], kwargs['bnb_dict']['bnb_weight'], bias=None, state=kwargs['bnb_dict']['state'])
    }

    kernelfns = {k: kernelfns[k] for k in kernels}


    triton_benchmark_sizes(inputs, kernelfns, sizes=hidden_sizes, metric_config=metric_config)

if __name__ == '__main__':
    lmetric_quant_matmul_Lx3HxH()
# %%



from .utils import download_hf_repo
from .floating_ops import lmetric_matmul_fp16
from .llm_quant import lmetric_quant_matmul_Lx3HxH
from .llm_prune import lmetric_prune_matmul_Lx3HxH
from .attention import lmetric_attention


import re
import multiprocessing
import itertools

split_str = 'llama.cpp: loading model from'

import argparse


tuning_args = [
    {'name': 'n_batch', 'type': int, 'default': [512], 'help': 'batch size'},
    {'name': 'n_gpu_layers', 'type': int, 'default': [0], 'help': 'number of gpu layers'},
    {'name': 'n_threads', 'type': int, 'default': [max(multiprocessing.cpu_count() // 2, 1)], 'help': 'number of threads'},
    {'name': 'last_n_tokens_size', 'type': int, 'default': [64], 'help': 'max number of tokens to keep in last_n_tokens deque'},
    {'name': 'use_mmap', 'type': bool, 'default': [True], 'help': 'use mmap'},
    {'name': 'use_mlock', 'type': bool, 'default': [False], 'help': 'use mlock'},
    {'name': 'low_vram', 'type': bool, 'default': [False], 'help': 'use low vram'},
]

def get_arg_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--model_path', required=True, type=str, help='model path')
    parser.add_argument('--prompt', '-p', type=str, default="Q: Name the planets in the solar system? A: ", help='input text')
    parser.add_argument('--max_tokens', type=int, default=32, help='max num of tokens of input text')

    for targs in tuning_args:
        parser.add_argument(f"--{targs['name']}", type=targs['type'], nargs='+', default=targs['default'], help=targs['help'])

    parser.add_argument('--n_repeats', type=int, default=1, help='number of repeatitions for benchmarking')
    return parser

def preprocess_args(args):
    vargs = vars(args)

    # Cmd
    cmd = [
            'python', '-m', 'lmetric.llamacpp.execute_llamacpp',
            '--model_path', args.model_path,
            '-p', args.prompt,
            '--max_tokens', str(args.max_tokens)
    ]

    for targs in tuning_args:
        if vargs[targs['name']] is not None:
            cmd += [f"--{targs['name']}"]
            cmd += [str(x) for x in vargs[targs['name']]]


    # Param
    model_path = args.model_path
    prompt = args.prompt
    max_tokens = args.max_tokens
    n_repeats = args.n_repeats

    vargs.pop('prompt')
    vargs.pop('model_path')
    vargs.pop('max_tokens')
    vargs.pop('n_repeats')
    param_combinations = list(itertools.product(*vargs.values()))
    params = []
    for combination in param_combinations:
        params_dict = {key: value for key, value in zip(vargs.keys(), combination)}
        params.append(params_dict)
    
    return vargs, model_path, prompt, max_tokens, n_repeats, cmd, params

def convert_to_json(array):
    result = {}

    load_time = re.search(r'load time =\s*([\d.]+)', array[0])
    if load_time:
        result['load_time'] = float(load_time.group(1))

    sample_time = re.search(r'sample time =\s*([\d.]+)', array[1])
    sample_runs = re.search(r'(\d+)\s+runs', array[1])
    if sample_time and sample_runs:
        result['sample_time'] = float(sample_time.group(1))
        result['sample_time_pt'] = result['sample_time'] / int(sample_runs.group(1))

    prompt_eval_time = re.search(r'prompt eval time =\s*([\d.]+)', array[2])
    prompt_eval_tokens = re.search(r'(\d+)\s+tokens', array[2])
    if prompt_eval_time and prompt_eval_tokens:
        result['prompt_eval_time'] = float(prompt_eval_time.group(1))
        result['prompt_eval_time_pt'] = result['prompt_eval_time'] / int(prompt_eval_tokens.group(1))

    eval_time = re.search(r'eval time =\s*([\d.]+)', array[3])
    eval_runs = re.search(r'(\d+)\s+runs', array[3])
    if eval_time and eval_runs:
        result['eval_time'] = float(eval_time.group(1))
        result['eval_time_pt'] = result['eval_time'] / int(eval_runs.group(1))

    total_time = re.search(r'total time =\s*([\d.]+)', array[4])
    if total_time:
        result['total_time'] = float(total_time.group(1))

    return result



# %%
import subprocess
import json
import triton
from tqdm import tqdm
import itertools

from lmetric.llamacpp.utils import convert_to_json, split_str, tuning_args, get_arg_parser, preprocess_args


args = get_arg_parser().parse_args()
vargs, model_path, prompt, max_tokens, n_repeats, cmd, params = preprocess_args(args)

# Run
measures = []
for _ in params:
    measures.append({
        'prompt_eval_max': 0,
        'prompt_eval_min': 1e8,
        'prompt_eval_avg': 0,
        'eval_max': 0,
        'eval_min': 1e8,
        'eval_avg': 0,
    })

for _ in tqdm(range(n_repeats)):
    result = subprocess.run(
        cmd,
        stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )

    decoded_str = result.stderr.decode('utf-8')
    results = decoded_str.split(split_str)[1:]

    for idx, res in enumerate(results):
        timing = [line for line in
                res.split('\n') if line.startswith('llama_print_timings')]
        timing = convert_to_json(timing)
        # print(json.dumps(convert_to_json(timing), indent=4))
        prompt_eval_time = timing['prompt_eval_time_pt']
        eval_time = timing['eval_time_pt']

        if prompt_eval_time > measures[idx]['prompt_eval_max']:
            measures[idx]['prompt_eval_max'] = prompt_eval_time
        if prompt_eval_time < measures[idx]['prompt_eval_min']:
            measures[idx]['prompt_eval_min'] = prompt_eval_time
        measures[idx]['prompt_eval_avg'] += prompt_eval_time / n_repeats
        if eval_time > measures[idx]['eval_max']:
            measures[idx]['eval_max'] = eval_time
        if eval_time < measures[idx]['eval_min']:
            measures[idx]['eval_min'] = eval_time
        measures[idx]['eval_avg'] += eval_time / n_repeats

# Search
min_avg_eval = 1e8
min_avg_eval_param = None
min_avg_peval = 1e8
min_avg_peval_param = None
for idx, measure in enumerate(measures):
    if measure['prompt_eval_avg'] < min_avg_peval:
        min_avg_peval = measure['prompt_eval_avg']
        min_avg_peval_param = params[idx]
    if measure['eval_avg'] < min_avg_eval:
        min_avg_eval = measure['eval_avg']
        min_avg_eval_param = params[idx]

print('===================')
print(f'min eval avg: {min_avg_eval}, min eval param: {min_avg_eval_param}')
print(f'min prompt eval avg: {min_avg_peval}, min prompt eval param: {min_avg_peval_param}')
print('===================')


# Plot
is_plotting = False
plotting_args = ''
dim_cnt = 0
for targs in tuning_args:
    val = vargs[targs['name']]
    if isinstance(val, list) and len(val) > 1:
        dim_cnt += 1
        plotting_args = targs['name']
if dim_cnt != 1:
    is_plotting = False
else:
    is_plotting = True

def get_param_idx_with_val(params, name, val):
    for idx, p in enumerate(params):
        if p[name] == val:
            return idx
    return -1

if is_plotting:
    # Plot
    @triton.testing.perf_report(
        triton.testing.Benchmark(
            x_names=['val'],  # Argument names to use as an x-axis for the plot.
            x_vals=vargs[plotting_args],  # Different possible values for `x_name`.
            x_log=False,  # x axis is logarithmic.
            line_arg='provider',  # Argument name whose value corresponds to a different line in the plot.
            line_vals=[''],  # Possible values for `line_arg`.
            line_names=[''],  # Label name for the lines.
            #styles=['-'],  # Line styles.
            ylabel='Prompt Eval Time',  # Label name for the y-axis.
            plot_name='Prompt Eval Time',  # Name for the plot. Used also as a file name for saving the plot.
            args={},  # Values for function arguments not in `x_names` and `y_name`.
        )
    )
    def benchmark1(val, provider):
        # input initialization
        idx = get_param_idx_with_val(params, plotting_args, val)
        assert idx != -1
        return measures[idx]['prompt_eval_avg'], \
                measures[idx]['prompt_eval_max'], \
                measures[idx]['prompt_eval_min']

    @triton.testing.perf_report(
        triton.testing.Benchmark(
            x_names=['val'],  # Argument names to use as an x-axis for the plot.
            x_vals=vargs[plotting_args],  # Different possible values for `x_name`.
            x_log=False,  # x axis is logarithmic.
            line_arg='provider',  # Argument name whose value corresponds to a different line in the plot.
            line_vals=[''],  # Possible values for `line_arg`.
            line_names=[''],  # Label name for the lines.
            #styles=['-'],  # Line styles.
            ylabel='Eval Time',  # Label name for the y-axis.
            plot_name='Eval Time',  # Name for the plot. Used also as a file name for saving the plot.
            args={},  # Values for function arguments not in `x_names` and `y_name`.
        )
    )
    def benchmark2(val, provider):
        # input initialization
        idx = get_param_idx_with_val(params, plotting_args, val)
        assert idx != -1
        return measures[idx]['eval_avg'], \
                measures[idx]['eval_max'], \
                measures[idx]['eval_min']

    benchmark1.run(print_data=True, show_plots=True)
    benchmark2.run(print_data=True, show_plots=True)
# %%


# %%
from llama_cpp import Llama

import sys
from typing import Optional, Union

from lmetric.llamacpp.utils import split_str, get_arg_parser, preprocess_args

args = get_arg_parser().parse_args()
vargs, model_path, prompt, max_tokens, n_repeats, cmd, params = preprocess_args(args)

for param in params:
    llm = Llama(
        model_path=model_path,
        **param
        )

    llm(prompt, max_tokens=max_tokens, stop=["Q:", "\n"], echo=False)

    del llm


# %%


# %%
import torch

import bitsandbytes as bnb
from bitsandbytes import functional as F

from pymeten import triton_benchmark_sizes, MetricConfig, MetricUnit
from scipy.stats import norm

# %%

def get_coo(x, threshold=0):
    """
    Returns a COO sparse tensor representation of the input tensor x based on a given threshold.

    Args:
        x (torch.Tensor): Input tensor.
        threshold (float, optional): Threshold value. Defaults to 0.

    Returns:
        torch.COOSparseTensor: COO sparse tensor representation of x.
    """

    #if threshold ==0:
    #    threshold = torch.quantile(x[:128][:2048].float(), ratio)
    #print(f'{ratio} {threshold}')
    idx = x <= threshold
    nnz = (idx == 1).sum().item()
    #print(f'nonzero ratio: {nnz / idx.numel()}')
    rows, cols = torch.where(idx)
    values = x[idx]
    cooA = F.COOSparseTensor(
        x.shape[0], x.shape[1], nnz, rows.int(), cols.int(), values
    )
    return cooA

def get_pt_sparse(x, threshold=0, representation='COO'):
    """
    Returns a PyTorch sparse tensor representation of the input tensor x based on a given threshold and representation type.

    Args:
        x (torch.Tensor): Input tensor.
        threshold (float, optional): Threshold value. Defaults to 0.
        representation (str, optional): Sparse tensor representation type. Defaults to 'COO'. options: COO CSR CSC BSR BSC

    Returns:
        torch.Tensor: Sparse tensor representation of x.
    """

    #if threshold ==0:
    #    threshold = torch.quantile(x[:128][:2048].float(), ratio)

    newx= torch.where(x>threshold, torch.tensor(0), x)
    if representation == 'COO':
        newx= newx.to_sparse_coo()
    elif representation == 'BSC':
        newx= newx.to_sparse_bsc()
    elif representation == 'BSR':
        newx= newx.to_sparse_bsr()
    elif representation == 'CSC':
        newx= newx.to_sparse_csc()
    elif representation == 'CSR':
        newx= newx.to_sparse_csr()
    else:
        newx= newx.to_sparse()
        print(type(newx))

    #print(newx)

    return newx

# %%
def lmetric_prune_matmul_Lx3HxH(seqlen=128, hidden_sizes=[2048 * i for i in range(1, 3)], dtype=torch.float16, metric='MS',
                                kernels=['torch', 'bnb_sparse'],
                                representations=['COO'],
                                ratios=[0.5]):
    """
    Computes metrics for pruning and matrix multiplication operations.

    Args:
        seqlen (int, optional): Sequence length. Defaults to 128.
        hidden_sizes (list, optional): List of hidden layer sizes. Defaults to [2048, 4096].
        dtype (torch.dtype, optional): Data type of the tensors. Defaults to torch.float16.
        metric (str, optional): Metric type ('TFLOPS', 'TBPS', or 'MS'). Defaults to 'MS'.
        kernels (list, optional): List of kernel types ('torch', 'bnb_sparse', 'torch_sparse'). Defaults to ['torch', 'bnb_sparse'].
        representations (list, optional): List of sparse tensor representations ('COO', 'CSC', 'CSR', 'BSC', 'BSR'). Defaults to ['COO'].
        ratios (list, optional): List of pruning ratios. Defaults to [0.5].
    """
    
    dtype_bytes = torch.finfo(dtype).bits // 8

    if metric == 'TFLOPS':
        metric_config=MetricConfig(
            MetricUnit.TFLOPS,
            calc_measure_fn=lambda size, ms: dtype_bytes*size*size*size / ms)
    elif metric == 'TBPS':
        metric_config=MetricConfig(
            MetricUnit.TBPS,
            calc_measure_fn=lambda size, ms: 3*dtype_bytes*size*size / ms)
    else:
        metric_config=MetricConfig(
            MetricUnit.MS)
        
    for k in kernels:
        assert k in ['torch', 'bnb_sparse', 'torch_sparse']
    for p in representations:
        assert p in ['COO', 'CSC', 'CSR', 'BSC', 'BSR']
    for r in ratios:
        assert r > 0 and r < 1
    if 'bnb_sparse' in kernels:
        assert dtype == torch.float16, 'bnb sparse only support float16'
    if 'torch_sparse' in kernels:
        assert dtype == torch.float32, 'torch sparse only support float32'
    
    inputs = {
        'x': lambda size: torch.randn(seqlen, 3*size, device='cuda', dtype=dtype),
        'w': lambda size: torch.randn(3*size, size, device='cuda', dtype=dtype),
        'wt': lambda size: torch.randn(size, 3*size, device='cuda', dtype=dtype),
    }

    for k in kernels:
        if k == 'bnb_sparse':
            for r in ratios:
                inputs.update({f'{k}_COO_{r}':lambda size:
                    get_coo(torch.randn(3*size, size, device='cuda', dtype=dtype), threshold=norm.ppf(r))
                })
        elif k == 'torch_sparse':
            for r in ratios:
                for p in representations:
                    inputs.update({f'{k}_{p}_{r}':lambda size:
                        get_pt_sparse(torch.randn(3*size, size, device='cuda', dtype=dtype), threshold=norm.ppf(r), representation=p)
                    })
                
    if 'torch' in kernels:
        kernelfns = {
            #'bnb.matmul': lambda **kwargs: 
            #    bnb.matmul(kwargs['x'], kwargs['wt']),
            'torch': lambda **kwargs: 
                torch.matmul(kwargs['x'], kwargs['w']),
        }
    else:
        kernelfns = {}

    for k in kernels:
        if k == 'bnb_sparse':
            for r in ratios:
                kernelfns.update({
                    f'{k}_COO_{r}': lambda **kwargs: 
                        F.spmm_coo(kwargs[f'{k}_COO_{r}'], kwargs['wt']),
                    
                })
        elif k == 'torch_sparse':
            for r in ratios:
                for p in representations:
                    kernelfns.update({f'{k}_{p}_{r}':lambda **kwargs:
                        torch.sparse.mm(kwargs[f'{k}_{p}_{r}'], kwargs['wt'])
                    })

    triton_benchmark_sizes(inputs, kernelfns, sizes=hidden_sizes, metric_config=metric_config)

# %%
if __name__ == '__main__':
    lmetric_prune_matmul_Lx3HxH()

# %%


from huggingface_hub import snapshot_download, hf_hub_download
import argparse

# https://huggingface.co/docs/huggingface_hub/guides/download

def download_hf_repo(repo_id, repo_type=None, filename=None, white_patterns=None, black_patterns=None,
                     save_dir=None, save_with_symlinks=False, revision=None):
    """
    Download a repository from Hugging Face Hub or create a snapshot.

    Args:
        repo_id (str): Identifier of the repository in the Hugging Face Hub.
        repo_type (str, optional): Type of the repository. Defaults to None.
        filename (str, optional): File name to download. Defaults to None.
        white_patterns (List[str], optional): List of patterns for including files. Defaults to None.
        black_patterns (List[str], optional): List of patterns for excluding files. Defaults to None.
        save_dir (str, optional): Directory to save the downloaded repository. Defaults to None.
        save_with_symlinks (bool, optional): Whether to use symlinks when saving locally. Defaults to False.
        revision (str, optional): Revision of the repository. Defaults to None.

    Returns:
        None

    Raises:
        SomeException: Description of the exception raised.

    Examples:
        
        download_hf_repo('NousResearch/llama2-7b-hf', save_dir='llama2-7b)

    """

    if filename is not None:
        hf_hub_download(repo_id=repo_id, repo_type=repo_type, filename=filename,
                        local_dir=save_dir, local_dir_use_symlinks=save_with_symlinks,
                        revision=revision)
    else:
        snapshot_download(repo_id=repo_id, repo_type=repo_type,
                          allow_patterns=white_patterns, ignore_patterns=black_patterns,
                          local_dir=save_dir, local_dir_use_symlinks=save_with_symlinks,
                          revision=revision)

if __name__ == '__main__':

    parser = argparse.ArgumentParser()

    parser.add_argument('--repo_id', required=True, type=str, help='repo id in hf')
    parser.add_argument('--filename', type=str, help='file name you want to download')
    parser.add_argument('--white_patterns', type=str, nargs='+', help='white file patterns')
    parser.add_argument('--black_patterns', type=str, nargs='+', help='black file patterns')
    parser.add_argument('--save_dir', type=str, help='save location if you do not want cache')
    parser.add_argument('--save_with_symlinks', type=str, help='whether use symlinks when saving locally')
    parser.add_argument('--revision', type=str, help='revision')

    args = parser.parse_args()
    kwargs = vars(args)

    download_hf_repo(**kwargs)


# %%
import math
import torch
import torch.nn.functional as F

from pymeten import triton_benchmark_sizes, MetricConfig, MetricUnit

from einops import rearrange

from typing import List

try:
    from flash_attn import flash_attn_qkvpacked_func
except:
    flash_attn_qkvpacked_func=None
from triton.ops.flash_attention import attention as attention_triton
try:
    import xformers.ops as xops
except:
    xops=None


torch.manual_seed(0)


# %%
def attention_pytorch(qkv, dropout_p=0.0, causal=True):
    """
    Arguments:
        qkv: (batch_size, seqlen, 3, nheads, head_dim)
        dropout_p: float
    Output:
        output: (batch_size, seqlen, nheads, head_dim)
    """
    batch_size, seqlen, _, nheads, d = qkv.shape
    q, k, v = qkv.unbind(dim=2)
    q = rearrange(q, 'b t h d -> (b h) t d')
    k = rearrange(k, 'b s h d -> (b h) d s')
    softmax_scale = 1.0 / math.sqrt(d)
    # Preallocate attn_weights for `baddbmm`
    scores = torch.empty(batch_size * nheads, seqlen, seqlen, dtype=qkv.dtype, device=qkv.device)
    scores = rearrange(torch.baddbmm(scores, q, k, beta=0, alpha=softmax_scale),
                       '(b h) t s -> b h t s', h=nheads)
    if causal:
        # "triu_tril_cuda_template" not implemented for 'BFloat16'
        # So we have to construct the mask in float
        causal_mask = torch.triu(torch.full((seqlen, seqlen), -10000.0, device=scores.device), 1)
        # TD [2022-09-30]: Adding is faster than masked_fill_ (idk why, just better kernel I guess)
        scores = scores + causal_mask.to(dtype=scores.dtype)
    attention = torch.softmax(scores, dim=-1)
    attention_drop = F.dropout(attention, dropout_p)
    output = torch.einsum('bhts,bshd->bthd', attention_drop , v)
    return output.to(dtype=qkv.dtype)


def lmetric_attention(batch_size: int or List[int]=32,
                                seqlen: int or List[int]=128,
                                num_heads: int or List[int]=[32 * i for i in range(1, 6)],
                                head_dim: int or List[int]=64,
                                dtype=torch.float16, metric='TFLOPS',
                                kernels=['torch', 'triton_flash', 'flash2', 'xformers'],
                                direction='fwd'
                                ):
    """
    Performs a attention module benchmarking with different kernels.

    Args:
        batch_size (int|list[int]): Batch size of the input data. Default is 32.
        seqlen (int|list[int]): Sequence length of the input data. Default is 128.
        num_heads (list[int]): Number of heads. Default is [32, 64, 96, 128, 160].
        head_dim (list[int]): The head dimension. Default is 64. triton_flash only support 64, and flash2 only support <256.
        dtype: Data type of the tensors. Default is torch.float16.
        metric (str): Metric to measure performance. Available options are 'TFLOPS', 'TBPS', or 'MS'. Default is 'TFLOPS'.
        kernels (list[str]): List of kernel functions to use. Default is ['torch', 'triton_flash', 'flash2', 'xformers'].
        direction (str): Direction of computation. Available options are 'fwd' (forward), 'bwd' (backward), or 'fwd_bwd'. Default is 'fwd'.

    Returns:
        None
    """
    
    dtype_bytes = torch.finfo(dtype).bits // 8
    
    comp_bs, comp_seqlen, comp_numhead, comp_headdim = False, False, False, False
    
    if isinstance(batch_size, List):
        comp_bs = True
    if isinstance(seqlen, List):
        comp_seqlen = True
    if isinstance(num_heads, List):
        comp_numhead = True
    if isinstance(head_dim, List):
        comp_headdim = True
    assert sum([comp_bs, comp_seqlen, comp_numhead, comp_headdim]) == 1, \
        "Must make one and only one of the following as a list: batch_size, \
        seqlen, num_heads, head_dim"

    if 'triton_flash' in kernels:
        assert head_dim == 64 or \
            isinstance(head_dim, List) and len(head_dim) == 1 and head_dim[0] == 64, \
                "triton_flash only support head_dim==64"
    
    def getTFLOPS(size, ms):
        if comp_bs:
            return dtype_bytes * size * seqlen**2 * num_heads * head_dim /ms
        if comp_seqlen:
            return dtype_bytes * batch_size * size**2 * num_heads * head_dim /ms
        if comp_numhead:
            return dtype_bytes * batch_size * seqlen**2 * size * head_dim /ms
        if comp_headdim:
            return dtype_bytes * batch_size * seqlen**2 * num_heads * size /ms

    def getTBPS(size, ms, is_forward=True):
        if comp_bs:
            if is_forward:
                return (4*dtype_bytes * size * seqlen * num_heads * head_dim +
                    (2*dtype_bytes+1) * size * seqlen**2 * num_heads * head_dim /64) / ms
            else:
                return (max(
                    2*dtype_bytes * size * seqlen * num_heads * head_dim,
                    2 * size * seqlen**2 * num_heads * head_dim/64) +
                    dtype_bytes * size * seqlen**2 * num_heads * head_dim/64) / ms
        if comp_seqlen:
            if is_forward:
                return (4*dtype_bytes * batch_size * size * num_heads * head_dim +
                    (2*dtype_bytes+1) * batch_size * size**2 * num_heads * head_dim /64) / ms
            else:
                return (max(
                    2*dtype_bytes * batch_size * size * num_heads * head_dim,
                    2 * batch_size * size**2 * num_heads * head_dim/64) +
                    dtype_bytes * batch_size * size**2 * num_heads * head_dim/64) / ms
        if comp_numhead:
            if is_forward:
                return (4*dtype_bytes * batch_size * seqlen * size * head_dim +
                    (2*dtype_bytes+1) * batch_size * seqlen**2 * size * head_dim /64) / ms
            else:
                return (max(
                    2*dtype_bytes * batch_size * seqlen * size * head_dim,
                    2 * batch_size * seqlen**2 * size * head_dim/64) +
                    dtype_bytes * batch_size * seqlen**2 * size * head_dim/64) / ms
        if comp_headdim:
            if is_forward:
                return (4*dtype_bytes * batch_size * seqlen * num_heads * size +
                    (2*dtype_bytes+1) * batch_size * seqlen**2 * num_heads * size /64) / ms
            else:
                return (max(
                    2*dtype_bytes * batch_size * seqlen * num_heads * size,
                    2 * batch_size * seqlen**2 * num_heads * size/64) +
                    dtype_bytes * batch_size * seqlen**2 * num_heads * size/64) / ms

    if metric == 'TFLOPS':
        metric_config=MetricConfig(
            MetricUnit.TFLOPS,
            calc_measure_fn=lambda size, ms:
                2*getTFLOPS(size, ms) if direction=='fwd' else \
                5*getTFLOPS(size, ms) if direction=='bwd' else \
                7*getTFLOPS(size, ms))
    elif metric == 'TBPS':
        metric_config=MetricConfig(
            MetricUnit.TBPS, # QKVO 8bsh, QK sm mask 5bs^2a
            # TODO: don't assume head_dim is 64
            calc_measure_fn=lambda size, ms:
                getTBPS(size, ms, True) if direction=='bwd' else \
                getTBPS(size, ms, False) if direction=='fwd' else \
                getTBPS(size, ms, True) + getTBPS(size, ms, False)
        )
    else:
        metric_config=MetricConfig(
            MetricUnit.MS)
    
    def get_qkv(size, separate=False):
        if comp_bs:
            if separate:
                return torch.randn(size, num_heads, seqlen, head_dim, device='cuda', dtype=dtype)
            else:
                return torch.randn(size, seqlen, 3, num_heads, head_dim, device='cuda', dtype=dtype)
        if comp_seqlen:
            if separate:
                return torch.randn(batch_size, num_heads, size, head_dim, device='cuda', dtype=dtype)
            else:
                return torch.randn(batch_size, size, 3, num_heads, head_dim, device='cuda', dtype=dtype)
        if comp_numhead:
            if separate:
                return torch.randn(batch_size, size, seqlen, head_dim, device='cuda', dtype=dtype)
            else:
                return torch.randn(batch_size, seqlen, 3, size, head_dim, device='cuda', dtype=dtype)
        if comp_headdim:
            if separate:
                return torch.randn(batch_size, num_heads, seqlen, size, device='cuda', dtype=dtype)
            else:
                return torch.randn(batch_size, seqlen, 3, num_heads, size, device='cuda', dtype=dtype)
        

    inputs = {
        'qkv' : lambda size:
            get_qkv(size, False),
        'q': lambda size:
            get_qkv(size, True),
        'k': lambda size:
            get_qkv(size, True),
        'v': lambda size:
            get_qkv(size, True),
    }

    kernelfns = {
        'torch': lambda **kwargs:
            attention_pytorch(kwargs['qkv'], causal=False),
        'triton_flash': lambda **kwargs:
            attention_triton(kwargs['q'], kwargs['k'], kwargs['v'], 64**(-0.5)),
        'flash2': lambda **kwargs:
            flash_attn_qkvpacked_func(kwargs['qkv']),
        'xformers': lambda **kwargs:
            xops.memory_efficient_attention(kwargs['q'], kwargs['k'], kwargs['v'],
                                            attn_bias=None,
                                            op=(xops.fmha.cutlass.FwOp, xops.fmha.cutlass.BwOp))
    }

    if flash_attn_qkvpacked_func is None:
        kernels.remove('flash2')
    if xops is None:
        kernels.remove('xformers')

    kernelfns = {k: kernelfns[k] for k in kernels}


    triton_benchmark_sizes(inputs,
                           kernelfns,
                           sizes=batch_size if comp_bs else \
                                seqlen if comp_seqlen else \
                                num_heads if comp_numhead else \
                                head_dim,
                           metric_config=metric_config)

if __name__ == '__main__':
    lmetric_attention(direction='fwd_bwd')
    #lmetric_attention(batch_size=[32 * i for i in range (1, 31, 5)], num_heads=32, direction='fwd_bwd')
    #lmetric_attention(seqlen=[128 * i for i in range (1, 9, 2)], num_heads=32, direction='fwd_bwd')
    #lmetric_attention(head_dim=[64 * i for i in range (1, 5)], kernels=['torch', 'flash2', 'xformers'], num_heads=32, direction='fwd_bwd')
# %%



# %%
import torch

import triton
import triton.language as tl

from pymeten import checkit
from pymeten import triton_benchmark_sizes, MetricConfig, MetricUnit


import torch

import triton
import triton.language as tl


# %%

# Triton Matmul Kernel Definition

# `triton.jit`'ed functions can be auto-tuned by using the `triton.autotune` decorator, which consumes:
#   - A list of `triton.Config` objects that define different configurations of
#       meta-parameters (e.g., `BLOCK_SIZE_M`) and compilation options (e.g., `num_warps`) to try
#   - An auto-tuning *key* whose change in values will trigger evaluation of all the
#       provided configs
@triton.autotune(
    configs=[
        #triton.Config({'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 256, 'BLOCK_SIZE_K': 64, 'GROUP_SIZE_M': 8}, num_stages=3, num_warps=8),
        triton.Config({'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 256, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 128, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 128, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 32, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 32, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=5, num_warps=2),
        triton.Config({'BLOCK_SIZE_M': 32, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=5, num_warps=2),
    ],
    key=['M', 'N', 'K'],
)
@triton.jit
def matmul_kernel(
    # Pointers to matrices
    a_ptr, b_ptr, c_ptr,
    # Matrix dimensions
    M, N, K,
    # The stride variables represent how much to increase the ptr by when moving by 1
    # element in a particular dimension. E.g. `stride_am` is how much to increase `a_ptr`
    # by to get the element one row down (A has M rows).
    stride_am, stride_ak,
    stride_bk, stride_bn,
    stride_cm, stride_cn,
    # Meta-parameters
    BLOCK_SIZE_M: tl.constexpr, BLOCK_SIZE_N: tl.constexpr, BLOCK_SIZE_K: tl.constexpr,
    GROUP_SIZE_M: tl.constexpr,
    ACTIVATION: tl.constexpr,
):
    """Kernel for computing the matmul C = A x B.
    A has shape (M, K), B has shape (K, N) and C has shape (M, N)
    """
    # -----------------------------------------------------------
    # Map program ids `pid` to the block of C it should compute.
    # This is done in a grouped ordering to promote L2 data reuse.
    # See above `L2 Cache Optimizations` section for details.
    pid = tl.program_id(axis=0)
    num_pid_m = tl.cdiv(M, BLOCK_SIZE_M)
    num_pid_n = tl.cdiv(N, BLOCK_SIZE_N)
    num_pid_in_group = GROUP_SIZE_M * num_pid_n
    group_id = pid // num_pid_in_group
    first_pid_m = group_id * GROUP_SIZE_M
    group_size_m = min(num_pid_m - first_pid_m, GROUP_SIZE_M)
    pid_m = first_pid_m + (pid % group_size_m)
    pid_n = (pid % num_pid_in_group) // group_size_m

    # Without GROUP_SIZE_M
    #num_pid_n = tl.cdiv(N, BLOCK_SIZE_N)
    #pid_m = pid // num_pid_n
    #pid_n = pid % num_pid_n

    # ----------------------------------------------------------
    # Create pointers for the first blocks of A and B.
    # We will advance this pointer as we move in the K direction
    # and accumulate
    # `a_ptrs` is a block of [BLOCK_SIZE_M, BLOCK_SIZE_K] pointers
    # `b_ptrs` is a block of [BLOCK_SIZE_K, BLOCK_SIZE_N] pointers
    # See above `Pointer Arithmetics` section for details
    offs_am = (pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)) % M
    offs_bn = (pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)) % N
    offs_k = tl.arange(0, BLOCK_SIZE_K)
    a_ptrs = a_ptr + (offs_am[:, None] * stride_am + offs_k[None, :] * stride_ak)
    b_ptrs = b_ptr + (offs_k[:, None] * stride_bk + offs_bn[None, :] * stride_bn)

    # -----------------------------------------------------------
    # Iterate to compute a block of the C matrix.
    # We accumulate into a `[BLOCK_SIZE_M, BLOCK_SIZE_N]` block
    # of fp32 values for higher accuracy.
    # `accumulator` will be converted back to fp16 after the loop.
    accumulator = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.float32)
    for k in range(0, tl.cdiv(K, BLOCK_SIZE_K)):
        # Load the next block of A and B, generate a mask by checking the K dimension.
        # If it is out of bounds, set it to 0.
        a = tl.load(a_ptrs, mask=offs_k[None, :] < K - k * BLOCK_SIZE_K, other=0.0)
        b = tl.load(b_ptrs, mask=offs_k[:, None] < K - k * BLOCK_SIZE_K, other=0.0)
        # We accumulate along the K dimension.
        accumulator += tl.dot(a, b)
        # Advance the ptrs to the next K block.
        a_ptrs += BLOCK_SIZE_K * stride_ak
        b_ptrs += BLOCK_SIZE_K * stride_bk
    # You can fuse arbitrary activation functions here
    # while the accumulator is still in FP32!
    if ACTIVATION == "leaky_relu":
        accumulator = leaky_relu(accumulator)
    c = accumulator.to(tl.float16)

    # -----------------------------------------------------------
    # Write back the block of the output matrix C with masks.
    offs_cm = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_cn = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    c_ptrs = c_ptr + stride_cm * offs_cm[:, None] + stride_cn * offs_cn[None, :]
    c_mask = (offs_cm[:, None] < M) & (offs_cn[None, :] < N)
    tl.store(c_ptrs, c, mask=c_mask)


# We can fuse `leaky_relu` by providing it as an `ACTIVATION` meta-parameter in `_matmul`.
@triton.jit
def leaky_relu(x):
    x = x + 1
    return tl.where(x >= 0, x, 0.01 * x)

def matmul(a, b, activation=""):
    # Check constraints.
    assert a.shape[1] == b.shape[0], "Incompatible dimensions"
    assert a.is_contiguous(), "Matrix A must be contiguous"
    assert b.is_contiguous(), "Matrix B must be contiguous"
    M, K = a.shape
    K, N = b.shape
    # Allocates output.
    c = torch.empty((M, N), device=a.device, dtype=a.dtype)
    # 1D launch kernel where each block gets its own program.
    grid = lambda META: (
        triton.cdiv(M, META['BLOCK_SIZE_M']) * triton.cdiv(N, META['BLOCK_SIZE_N']),
    )
    matmul_kernel[grid](
        a, b, c,
        M, N, K,
        a.stride(0), a.stride(1),
        b.stride(0), b.stride(1),
        c.stride(0), c.stride(1),
        ACTIVATION=activation
    )
    return c

# %%

def lmetric_matmul_fp16(sizes=[128 * i for i in range(2, 33, 5)], metric='TFLOPS', kernels=['triton', 'torch']):
    """
    Performs matrix multiplication using Triton and Torch libraries with FP16 data type.

    Args:
        sizes (list): List of sizes for the matrices. Default is [256, 896, 1536, 2176, 2816, 3456, 4096].
        metric (str): Metric to measure performance. Valid options are 'TFLOPS' (default), 'TBPS', and 'MS'.
        kernels (list): The kernels we are testing. Valid options are 'triton', 'torch'

    Returns:
        None

    """

    dtype = torch.float16
    if metric == 'TFLOPS':
        metric_config=MetricConfig(
            MetricUnit.TFLOPS,
            calc_measure_fn=lambda size, ms: 2*size*size*size / ms)
    elif metric == 'TBPS':
        metric_config=MetricConfig(
            MetricUnit.TBPS,
            calc_measure_fn=lambda size, ms: 3*2*size*size / ms)
    else:
        metric_config=MetricConfig(
            MetricUnit.MS)

    # Correctness Check

    print('--- correctness check with size 512x512 ---')
    torch.manual_seed(0)
    a = torch.randn((512, 512), device='cuda', dtype=dtype)
    b = torch.randn((512, 512), device='cuda', dtype=dtype)
    triton_output = matmul(a, b)
    torch_output = torch.matmul(a, b)

    checkit(lambda: torch.allclose(triton_output, torch_output))


    # Speed Test

    print('--- speed test ---')

    inputs = {
        'x': lambda size: torch.randn((size, size), device='cuda', dtype=dtype),
        'y': lambda size: torch.randn((size, size), device='cuda', dtype=dtype)
    }
    kernelfns = {
        'triton': lambda x, y: matmul(x, y),
        'torch': lambda x, y: torch.matmul(x, y)
    }

    kernelfns = {k: kernelfns[k] for k in kernels}

    triton_benchmark_sizes(inputs, kernelfns, sizes=sizes, metric_config=metric_config)

if __name__ == '__main__':
    lmetric_matmul_fp16()
# %%



from setuptools import setup, find_packages

VERSION = '0.0.4'
DESCRIPTION = 'Large Model Metrics'
LONG_DESCRIPTION = 'A tool for measuring training/inference of large models'

requirements = [
    'huggingface_hub',
    'triton',
    'scipy',
    'bitsandbytes',
    'pymeten',
    'einops',
]
devRequirements = [
    'sphinx',
    'sphinx_rtd_theme',
    'pytest',
]
attentionRequirements = [
    'flash-attn',
    'xformers',
]

setup(
    name="lmetric",
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    author="deciding",
    author_email="zhangzn710@gmail.com",
    license='MIT',
    packages=find_packages(),
    install_requires=requirements,
    extras_require={
        'dev': devRequirements,
        'attention': attentionRequirements
    },
    keywords='conversion',
    classifiers= [
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        'License :: OSI Approved :: MIT License',
        "Programming Language :: Python :: 3",
    ]
)
